# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - DateSupport library (derived from EventAggregatorSupport)

    @copyright: 2008, 2009, 2010, 2011, 2012 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

import calendar
import datetime
import re
import bisect

try:
    import pytz
except ImportError:
    pytz = None

__version__ = "0.1"

# Date labels.

month_labels = ["January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"]
weekday_labels = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

# Month, date, time and datetime parsing.

month_regexp_str = ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})'
date_regexp_str = ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
time_regexp_str = ur'(?P<hour>[0-2][0-9]):(?P<minute>[0-5][0-9])(?::(?P<second>[0-6][0-9]))?'
timezone_offset_str = ur'(?P<offset>(UTC)?(?:(?P<sign>[-+])?(?P<hours>[0-9]{2})(?::?(?P<minutes>[0-9]{2}))?))'
timezone_olson_str = ur'(?P<olson>[a-zA-Z]+(?:/[-_a-zA-Z]+){1,2})'
timezone_utc_str = ur'UTC'
timezone_regexp_str = ur'(?P<zone>' + timezone_offset_str + '|' + timezone_olson_str + '|' + timezone_utc_str + ')'
datetime_regexp_str = date_regexp_str + ur'(?:\s+' + time_regexp_str + ur'(?:\s+' + timezone_regexp_str + ur')?)?'

month_regexp = re.compile(month_regexp_str, re.UNICODE)
date_regexp = re.compile(date_regexp_str, re.UNICODE)
time_regexp = re.compile(time_regexp_str, re.UNICODE)
timezone_olson_regexp = re.compile(timezone_olson_str, re.UNICODE)
timezone_offset_regexp = re.compile(timezone_offset_str, re.UNICODE)
datetime_regexp = re.compile(datetime_regexp_str, re.UNICODE)

# iCalendar date and datetime parsing.

date_icalendar_regexp_str = ur'(?P<year>[0-9]{4})(?P<month>[0-9]{2})(?P<day>[0-9]{2})'
datetime_icalendar_regexp_str = date_icalendar_regexp_str + \
    ur'(?:' \
    ur'T(?P<hour>[0-2][0-9])(?P<minute>[0-5][0-9])(?P<second>[0-6][0-9])' \
    ur'(?P<utc>Z)?' \
    ur')?'

date_icalendar_regexp = re.compile(date_icalendar_regexp_str, re.UNICODE)
datetime_icalendar_regexp = re.compile(datetime_icalendar_regexp_str, re.UNICODE)

# Utility functions.

def sign(x):
    if x < 0:
        return -1
    else:
        return 1

def int_or_none(x):
    if x is None:
        return x
    else:
        return int(x)

def getMonthLabel(month):

    "Return an unlocalised label for the given 'month'."

    return month_labels[month - 1] # zero-based labels

def getDayLabel(weekday):

    "Return an unlocalised label for the given 'weekday'."

    return weekday_labels[weekday]

# Interfaces.

class ActsAsTimespan:
    pass

# Date-related functions.

def cmp_dates_as_day_start(a, b):

    """
    Compare dates/datetimes 'a' and 'b' treating dates without time information
    as the earliest time in a particular day.
    """

    are_equal = a == b

    if are_equal:
        a2 = a.as_datetime_or_date()
        b2 = b.as_datetime_or_date()

        if isinstance(a2, Date) and isinstance(b2, DateTime):
            return -1
        elif isinstance(a2, DateTime) and isinstance(b2, Date):
            return 1

    return cmp(a, b)

class Convertible:

    "Support for converting temporal objects."

    def _get_converter(self, resolution):
        if resolution == "month":
            return lambda x: x and x.as_month()
        elif resolution == "date":
            return lambda x: x and x.as_date()
        elif resolution == "datetime":
            return lambda x: x and x.as_datetime_or_date()
        else:
            return lambda x: x

class Temporal(Convertible):

    "A simple temporal representation, common to dates and times."

    def __init__(self, data):
        self.data = list(data)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.data)

    def __hash__(self):
        return hash(self.as_tuple())

    def as_tuple(self):
        return tuple(self.data)

    def convert(self, resolution):
        return self._get_converter(resolution)(self)

    def __cmp__(self, other):

        """
        The result of comparing this instance with 'other' is derived from a
        comparison of the instances' date(time) data at the highest common
        resolution, meaning that if a date is compared to a datetime, the
        datetime will be considered as a date. Thus, a date and a datetime
        referring to the same date will be considered equal.
        """

        if not isinstance(other, Temporal):
            return NotImplemented
        else:
            data = self.as_tuple()
            other_data = other.as_tuple()
            length = min(len(data), len(other_data))
            return cmp(data[:length], other_data[:length])

    def __sub__(self, other):

        """
        Return the difference between this object and the 'other' object at the
        highest common accuracy of both objects.
        """

        if not isinstance(other, Temporal):
            return NotImplemented
        else:
            data = self.as_tuple()
            other_data = other.as_tuple()
            direction = self < other and 1 or -1

            if len(data) < len(other_data):
                return (len(self.until(other)) - 1) * direction
            else:
                return (len(other.until(self)) - 1) * -direction

    def _until(self, start, end, nextfn, prevfn):

        """
        Return a collection of units of time by starting from the given 'start'
        and stepping across intervening units until 'end' is reached, using the
        given 'nextfn' and 'prevfn' to step from one unit to the next.
        """

        current = start
        units = [current]
        if current < end:
            while current < end:
                current = nextfn(current)
                units.append(current)
        elif current > end:
            while current > end:
                current = prevfn(current)
                units.append(current)
        return units

    def ambiguous(self):

        "Only times can be ambiguous."

        return 0

class Month(Temporal):

    "A simple year-month representation."

    def __str__(self):
        return "%04d-%02d" % self.as_tuple()[:2]

    def as_datetime(self, day, hour, minute, second, zone):
        return DateTime(self.as_tuple() + (day, hour, minute, second, zone))

    def as_date(self, day):
        if day < 0:
            weekday, ndays = self.month_properties()
            day = ndays + 1 + day
        return Date(self.as_tuple() + (day,))

    def as_month(self):
        return self

    def year(self):
        return self.data[0]

    def month(self):
        return self.data[1]

    def month_properties(self):

        """
        Return the weekday of the 1st of the month, along with the number of
        days, as a tuple.
        """

        year, month = self.as_tuple()[:2]
        return calendar.monthrange(year, month)

    def month_update(self, n=1):

        "Return the month updated by 'n' months."

        year, month = self.as_tuple()[:2]
        return Month((year + (month - 1 + n) / 12, (month - 1 + n) % 12 + 1))

    update = month_update

    def next_month(self):

        "Return the month following this one."

        return self.month_update(1)

    next = next_month

    def previous_month(self):

        "Return the month preceding this one."

        return self.month_update(-1)

    previous = previous_month

    def months_until(self, end):

        "Return the collection of months from this month until 'end'."

        return self._until(self.as_month(), end.as_month(), Month.next_month, Month.previous_month)

    until = months_until

class Date(Month):

    "A simple year-month-day representation."

    def constrain(self):
        year, month, day = self.as_tuple()[:3]

        month = max(min(month, 12), 1)
        wd, last_day = calendar.monthrange(year, month)
        day = max(min(day, last_day), 1)

        self.data[1:3] = month, day

    def __str__(self):
        return "%04d-%02d-%02d" % self.as_tuple()[:3]

    def as_datetime(self, hour, minute, second, zone):
        return DateTime(self.as_tuple() + (hour, minute, second, zone))

    def as_start_of_day(self):
        return self.as_datetime(None, None, None, None)

    def as_date(self):
        return self

    def as_datetime_or_date(self):
        return self

    def as_month(self):
        return Month(self.data[:2])

    def day(self):
        return self.data[2]

    def day_update(self, n=1):

        "Return the month updated by 'n' days."

        delta = datetime.timedelta(n)
        dt = datetime.date(*self.as_tuple()[:3])
        dt_new = dt + delta
        return Date((dt_new.year, dt_new.month, dt_new.day))

    update = day_update

    def next_day(self):

        "Return the date following this one."

        year, month, day = self.as_tuple()[:3]
        _wd, end_day = calendar.monthrange(year, month)
        if day == end_day:
            if month == 12:
                return Date((year + 1, 1, 1))
            else:
                return Date((year, month + 1, 1))
        else:
            return Date((year, month, day + 1))

    next = next_day

    def previous_day(self):

        "Return the date preceding this one."

        year, month, day = self.as_tuple()[:3]
        if day == 1:
            if month == 1:
                return Date((year - 1, 12, 31))
            else:
                _wd, end_day = calendar.monthrange(year, month - 1)
                return Date((year, month - 1, end_day))
        else:
            return Date((year, month, day - 1))

    previous = previous_day

    def days_until(self, end):

        "Return the collection of days from this date until 'end'."

        return self._until(self.as_date(), end.as_date(), Date.next_day, Date.previous_day)

    until = days_until

class DateTime(Date):

    "A simple date plus time representation."

    def constrain(self):
        Date.constrain(self)

        hour, minute, second = self.as_tuple()[3:6]

        if self.has_time():
            hour = max(min(hour, 23), 0)
            minute = max(min(minute, 59), 0)

            if second is not None:
                second = max(min(second, 60), 0) # support leap seconds

            self.data[3:6] = hour, minute, second

    def __str__(self):
        return Date.__str__(self) + self.time_string()

    def time_string(self, zone_as_offset=0):
        if self.has_time():
            data = self.as_tuple()
            time_str = " %02d:%02d" % data[3:5]
            if data[5] is not None:
                time_str += ":%02d" % data[5]
            if data[6] is not None:
                if zone_as_offset:
                    utc_offset = self.utc_offset()
                    if utc_offset:
                        time_str += " %+03d:%02d" % utc_offset
                else:
                    time_str += " %s" % data[6]
            return time_str
        else:
            return ""

    def as_HTTP_datetime_string(self):
        weekday = calendar.weekday(*self.data[:3])
        return "%s, %02d %s %04d %02d:%02d:%02d GMT" % ((
            getDayLabel(weekday),
            self.data[2],
            getMonthLabel(self.data[1]),
            self.data[0]
            ) + tuple(self.data[3:6]))

    def as_ISO8601_datetime_string(self):
        return Date.__str__(self) + self.time_string(1)

    def as_datetime(self):
        return self

    def as_date(self):
        return Date(self.data[:3])

    def as_datetime_or_date(self):

        """
        Return a date for this datetime if fields are missing. Otherwise, return
        this datetime itself.
        """

        if not self.has_time():
            return self.as_date()
        else:
            return self

    def __cmp__(self, other):

        """
        The result of comparing this instance with 'other' is, if both instances
        are datetime instances, derived from a comparison of the datetimes
        converted to UTC. If one or both datetimes cannot be converted to UTC,
        the datetimes are compared using the basic temporal comparison which
        compares their raw time data.
        """

        this = self

        if this.has_time():
            if isinstance(other, DateTime):
                if other.has_time():
                    this_utc = this.to_utc()
                    other_utc = other.to_utc()
                    if this_utc is not None and other_utc is not None:
                        return cmp(this_utc.as_tuple(), other_utc.as_tuple())
                else:
                    other = other.padded()
        else:
            this = this.padded()

        return Date.__cmp__(this, other)

    def __sub__(self, other):

        """
        Return the difference between this object and the 'other' object at the
        highest common accuracy of both objects.
        """

        if not isinstance(other, Temporal):
            return NotImplemented
        elif not other.has_time():
            return self.as_date() - other
        else:
            utc = self.to_utc()
            other = other.to_utc()
            days = utc.as_date() - other.as_date()
            h1, m1, s1 = utc.as_tuple()[3:6]
            h2, m2, s2 = other.as_tuple()[3:6]
            return days * 24 * 3600 + (h1 - h2) * 3600 + (m1 - m2) * 60 + s1 - s2

    def has_time(self):

        """
        Return whether this object has any time information. Objects without
        time information can refer to the very start of a day.
        """

        return self.data[3] is not None and self.data[4] is not None

    def time(self):
        return self.data[3:]

    def seconds(self):
        return self.data[5]

    def time_zone(self):
        return self.data[6]

    def set_time_zone(self, value):
        self.data[6] = value

    def padded(self, empty_value=0):

        """
        Return a datetime with missing fields defined as being the given
        'empty_value' or 0 if not specified.
        """

        data = []
        for x in self.data[:6]:
            if x is None:
                data.append(empty_value)
            else:
                data.append(x)

        data += self.data[6:]
        return DateTime(data)

    def to_utc(self):

        """
        Return this object converted to UTC, or None if such a conversion is not
        defined.
        """

        if not self.has_time():
            return None

        offset = self.utc_offset()
        if offset:
            hours, minutes = offset

            # Invert the offset to get the correction.

            hours, minutes = -hours, -minutes

            # Get the components.

            hour, minute, second, zone = self.time()
            date = self.as_date()

            # Add the minutes and hours.

            minute += minutes
            if minute < 0 or minute > 59:
                hour += minute / 60
                minute = minute % 60

            # NOTE: This makes various assumptions and probably would not work
            # NOTE: for general arithmetic.

            hour += hours
            if hour < 0:
                date = date.previous_day()
                hour += 24
            elif hour > 23:
                date = date.next_day()
                hour -= 24

            return date.as_datetime(hour, minute, second, "UTC")

        # Cannot convert.

        else:
            return None

    def utc_offset(self):

        "Return the UTC offset in hours and minutes."

        zone = self.time_zone()
        if not zone:
            return None

        # Support explicit UTC zones.

        if zone == "UTC":
            return 0, 0

        # Attempt to return a UTC offset where an explicit offset has been set.

        match = timezone_offset_regexp.match(zone)
        if match:
            if match.group("sign") == "-":
                offset_sign = -1
            else:
                offset_sign = 1

            hours = int(match.group("hours")) * offset_sign
            minutes = int(match.group("minutes") or 0) * offset_sign
            return hours, minutes

        # Attempt to handle Olson time zone identifiers.

        dt = self.as_olson_datetime()
        if dt:
            seconds = dt.utcoffset().seconds + dt.utcoffset().days * 24 * 3600
            hours = abs(seconds) / 3600
            minutes = (abs(seconds) % 3600) / 60
            return sign(seconds) * hours, sign(seconds) * minutes

        # Otherwise return None.

        return None

    def olson_identifier(self):

        "Return the Olson identifier from any zone information."

        zone = self.time_zone()
        if not zone:
            return None

        # Attempt to match an identifier.

        match = timezone_olson_regexp.match(zone)
        if match:
            return match.group("olson")
        else:
            return None

    def _as_olson_datetime(self, hours=None):

        """
        Return a Python datetime object for this datetime interpreted using any
        Olson time zone identifier and the given 'hours' offset, raising one of
        the pytz exceptions in case of ambiguity.
        """

        olson = self.olson_identifier()
        if olson and pytz:
            tz = pytz.timezone(olson)
            data = self.padded().as_tuple()[:6]
            dt = datetime.datetime(*data)

            # With an hours offset, find a time probably in a previously
            # applicable time zone.

            if hours is not None:
                td = datetime.timedelta(0, hours * 3600)
                dt += td

            ldt = tz.localize(dt, None)

            # With an hours offset, adjust the time to define it within the
            # previously applicable time zone but at the presumably intended
            # position.

            if hours is not None:
                ldt -= td

            return ldt
        else:
            return None

    def as_olson_datetime(self):

        """
        Return a Python datetime object for this datetime interpreted using any
        Olson time zone identifier, choosing the time from the zone before the
        period of ambiguity.
        """

        try:
            return self._as_olson_datetime()
        except (pytz.UnknownTimeZoneError, pytz.AmbiguousTimeError):

            # Try again, using an earlier local time and then stepping forward
            # in the chosen zone.
            # NOTE: Four hours earlier seems reasonable.

            return self._as_olson_datetime(-4)

    def ambiguous(self):

        "Return whether the time is local and ambiguous."

        try:
            self._as_olson_datetime()
        except (pytz.UnknownTimeZoneError, pytz.AmbiguousTimeError):
            return 1

        return 0

class Timespan(ActsAsTimespan, Convertible):

    """
    A period of time which can be compared against others to check for overlaps.
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

        # NOTE: Should perhaps catch ambiguous time problems elsewhere.

        if self.ambiguous() and self.start is not None and self.end is not None and start > end:
            self.start, self.end = end, start

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.start, self.end)

    def __hash__(self):
        return hash((self.start, self.end))

    def as_timespan(self):
        return self

    def as_limits(self):
        return self.start, self.end

    def ambiguous(self):
        return self.start is not None and self.start.ambiguous() or self.end is not None and self.end.ambiguous()

    def convert(self, resolution):
        return Timespan(*map(self._get_converter(resolution), self.as_limits()))

    def is_before(self, a, b):

        """
        Return whether 'a' is before 'b'. Since the end datetime of one period
        may be the same as the start datetime of another period, and yet the
        first period is intended to be concluded by the end datetime and not
        overlap with the other period, a different test is employed for datetime
        comparisons.
        """

        # Datetimes without times can be equal to dates and be considered as
        # occurring before those dates. Generally, datetimes should not be
        # produced without time information as getDateTime converts such
        # datetimes to dates.

        if isinstance(a, DateTime) and (isinstance(b, DateTime) or not a.has_time()):
            return a <= b
        else:
            return a < b

    def __contains__(self, other):

        """
        This instance is considered to contain 'other' if one is not before or
        after the other. If this instance overlaps or coincides with 'other',
        then 'other' is regarded as belonging to this instance's time period.
        """

        return self == other

    def __cmp__(self, other):

        """
        Return whether this timespan occupies the same period of time as the
        'other'. Timespans are considered less than others if their end points
        precede the other's start point, and are considered greater than others
        if their start points follow the other's end point.
        """

        if isinstance(other, ActsAsTimespan):
            other = other.as_timespan()

            if self.end is not None and other.start is not None and self.is_before(self.end, other.start):
                return -1
            elif self.start is not None and other.end is not None and self.is_before(other.end, self.start):
                return 1
            else:
                return 0

        else:
            if self.end is not None and self.is_before(self.end, other):
                return -1
            elif self.start is not None and self.is_before(other, self.start):
                return 1
            else:
                return 0

class TimespanCollection:

    """
    A class providing a list-like interface supporting membership tests at a
    particular resolution in order to maintain a collection of non-overlapping
    timespans.
    """

    def __init__(self, resolution, values=None):
        self.resolution = resolution
        self.values = values or []

    def as_timespan(self):
        return Timespan(*self.as_limits())

    def as_limits(self):

        "Return the earliest and latest points in time for this collection."

        if not self.values:
            return None, None
        else:
            first, last = self.values[0], self.values[-1]
            if isinstance(first, ActsAsTimespan):
                first = first.as_timespan().start
            if isinstance(last, ActsAsTimespan):
                last = last.as_timespan().end
            return first, last

    def convert(self, value):
        if isinstance(value, ActsAsTimespan):
            ts = value.as_timespan()
            return ts and ts.convert(self.resolution)
        else:
            return value.convert(self.resolution)

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __getitem__(self, i):
        return self.values[i]

    def __setitem__(self, i, value):
        self.values[i] = value

    def __contains__(self, value):
        test_value = self.convert(value)
        return test_value in self.values

    def append(self, value):
        self.values.append(value)

    def insert(self, i, value):
        self.values.insert(i, value)

    def pop(self):
        return self.values.pop()

    def insert_in_order(self, value):
        bisect.insort_left(self, value)

def getDate(s):

    "Parse the string 's', extracting and returning a date object."

    dt = getDateTime(s)
    if dt is not None:
        return dt.as_date()
    else:
        return None

def getDateTime(s):

    """
    Parse the string 's', extracting and returning a datetime object where time
    information has been given or a date object where time information is
    absent.
    """

    m = datetime_regexp.search(s)
    if m:
        groups = list(m.groups())

        # Convert date and time data to integer or None.

        return DateTime(map(int_or_none, groups[:6]) + [m.group("zone")]).as_datetime_or_date()
    else:
        return None

def getDateFromCalendar(s):

    """
    Parse the iCalendar format string 's', extracting and returning a date
    object.
    """

    dt = getDateTimeFromCalendar(s)
    if dt is not None:
        return dt.as_date()
    else:
        return None

def getDateTimeFromCalendar(s):

    """
    Parse the iCalendar format datetime string 's', extracting and returning a
    datetime object where time information has been given or a date object where
    time information is absent.
    """

    m = datetime_icalendar_regexp.search(s)
    if m:
        groups = list(m.groups())

        # Convert date and time data to integer or None.

        return DateTime(map(int_or_none, groups[:6]) + [m.group("utc") and "UTC" or None]).as_datetime_or_date()
    else:
        return None

def getDateStrings(s):

    "Parse the string 's', extracting and returning all date strings."

    start = 0
    m = date_regexp.search(s, start)
    l = []
    while m:
        l.append("-".join(m.groups()))
        m = date_regexp.search(s, m.end())
    return l

def getMonth(s):

    "Parse the string 's', extracting and returning a month object."

    m = month_regexp.search(s)
    if m:
        return Month(map(int, m.groups()))
    else:
        return None

def getCurrentDate():

    "Return the current date as a (year, month, day) tuple."

    today = datetime.date.today()
    return Date((today.year, today.month, today.day))

def getCurrentMonth():

    "Return the current month as a (year, month) tuple."

    today = datetime.date.today()
    return Month((today.year, today.month))

def getCurrentYear():

    "Return the current year."

    today = datetime.date.today()
    return today.year

# vim: tabstop=4 expandtab shiftwidth=4
